'''
Module for fault tolerance handling using Finite State Machines (FSM)

Circuit breakers prevent additional load from being cast on an ailing systems
and provide graceful degradation of services to consumers of the ailing system.
They function by switching between 3 states:
    Closed:    Healthy normal state
    Open:      Unhealthy state, do not allow any traffic through
    Half Open: Send one test request through, and open or close based
               on the result of the test

Example Use:
    @CircuitBreaker(allowed_fails=2, retry_time=10, retry_after=50)
    def request_to_service():
        # send http request to other system...

    If exceptions are encountered by request_to_service, they will increment
    the internal failure counter of CircuitBreaker until it hits the
    allowed_fails specified by the user. At this point the breaker will
    spring open and not let anymore requests occur until the breaker is
    set to the half open state when retry_time has passed in seconds or
    number of requests on open circuit crosses retry_after (whatever happens
    first)

'''
import time
import functools
import threading
from config import logger
from requests.exceptions import ConnectionError

CLOSED = 0
OPEN = 1
HALF_OPEN = 2


class CircuitBreaker(object):
    '''FSM to allow fault tolerance when systems fail'''
    def __init__(self, allowed_fails=3, retry_time=30, retry_after=50,
                 validation_func=None, allowed_exceptions=None,
                 failure_exceptions=None):
        '''
        Initializes Breaker object

        Args:
            allowed_fails(int): Maximum number of consecutive failures allowed
                before opening circuit
            retry_time(int): Number of seconds during close period before
                allowing test request to check if other end of circuit is
                responsive
            retry_after(int): Number of max failed requests on open circuit
                after which the circuit will be half_open irrespective of
                retry_time
            validation_func(func): function to check if return value of wrapped
                function is permissible. Must return boolean value
                allowed_exceptions(list[Exception]): permissible exceptions
                that will not trigger a failure. Do not use in conjunction with
                failure_exceptions. Will also check for child exceptions of the
                ones provided here. If these exceptions are caught, they will
                not be counted as a success either, and will not change the
                state of the FSM
            failure_exceptions(list[Exception]): if provided, only these
                exceptions will be registered as failures. Do not use in
                conjunction with allowed_exceptions. Will also check for child
                exceptions of the ones provided
        '''
        self._allowed_fails = allowed_fails
        self._retry_time = retry_time
        self._validation_func = validation_func
        self._lock = threading.Lock()
        self._read_lock = threading.RLock()
        self._failure_count = 0
        self._open_circuit_failure_count = 0
        self._open_circuit_failure_threashold = retry_after
        self._state = CLOSED
        self._half_open_time = 0  # initialize to minimum seconds since epoch
        if allowed_exceptions is not None:
            self._allowed_exceptions = tuple(allowed_exceptions)
        else:
            self._allowed_exceptions = ()

        if failure_exceptions is not None:
            self._failure_exceptions = tuple(failure_exceptions)
        else:
            self._failure_exceptions = ()

        if self._failure_exceptions and self._allowed_exceptions:
            raise ValueError("Cannot set failure exceptions in tandem with "
                             "allowed_exceptions")

    def _open(self):
        '''Open the circuit breaker and set time for half open'''
        self._state = OPEN
        open_time = time.time()
        self._half_open_time = open_time + self._retry_time
        self._open_circuit_failure_count = 0
        logger.info("Circuit breaker opened")

    def _close(self):
        '''Close circuit breaker and reset failure count'''
        self._state = CLOSED
        self._failure_count = 0
        self._open_circuit_failure_count = 0
        logger.info("Circuit breaker closing, reset failure count")

    def _half_open(self):
        ''' Set circuit breaker to half open state'''
        self._state = HALF_OPEN
        logger.info("Circuit breaker half open")

    def _check_state(self):
        '''Check current state of breaker and set half open when possible'''
        if self._state == OPEN:
            now = time.time()
            if now >= self._half_open_time or \
                self._open_circuit_failure_count >= \
                    self._open_circuit_failure_threashold:
                self._half_open()

        return self._state

    def _on_failure(self):
        '''
        Increments failure counter and switches state if allowed_fails is
        reached
        '''
        self._failure_count += 1
        logger.debug("Failure encountered, failure count: {}".
                     format(self._failure_count))
        if self._failure_count >= self._allowed_fails:
            current_state = self._check_state()
            if current_state != OPEN:
                self._open()

    def _on_success(self):
        '''
        Moves breaker to closed state
        '''
        self._close()

    def _parse_result(self, result):
        '''
        Determine if result of wrapped function is valid

        Args:
            result(object): return value of wrapped function
        '''
        if self._validation_func is None:
            self._on_success()
            return

        if self._validation_func(result):
            self._on_success()
        else:
            self._on_failure()

    def _call(self, func, *args, **kwargs):
        '''
        Wraps decorated function and watches for successes and failures

        Args:
            func(function): decorated function
            *args: args passed to decorated function
            **kwargs: kwargs passed to decorated function
        '''
        with self._lock:
            current_state = self._check_state()
            if current_state == OPEN:
                self._open_circuit_failure_count += 1
                raise ConnectionError("Open circuit")
            try:
                result = func(*args, **kwargs)
            except self._allowed_exceptions as e:
                logger.info("Encountered allowed exception {}".
                            format(e.__class__))
                return  # not a failure, but not a success
            except self._failure_exceptions:
                logger.exception("Caught pre-defined failure exception")
                self._on_failure()
            except Exception as e:
                logger.exception("Caught unhandled exception, incrementing "
                                 "failure count")
                if self._failure_exceptions:
                    logger.info("Encountered non-failure exception {}".
                                format(e.__class__))
                    return  # not a failure, but not a success
                else:
                    self._on_failure()
            else:
                logger.debug("Successfully completed wrapped function")
                self._parse_result(result)

    def __call__(self, func):
        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            return self._call(func, *args, **kwargs)

        return wrapped_func

    @property
    def is_open(self):
        with self._lock:
            return self._check_state() == OPEN

    def increment_failure_count(self):
        with self._lock:
            self._open_circuit_failure_count += 1

    def mark_failure(self):
        with self._lock:
            self._on_failure()

    def close(self):
        with self._lock:
            if self._check_state() != CLOSED:
                self._close()

    def __repr__(self):
        with self._read_lock:
            return ("Circuit Breaker - state: {state} fails: {fails} allowed "
                    "fails: {allowed} retry time: {retry_time} retry after: "
                    "{retry_after} failed requests".format(
                        state=self._state,
                        fails=self._failure_count,
                        allowed=self._allowed_fails,
                        retry_time=self._retry_time,
                        retry_after=self._open_circuit_failure_threashold)
                    )



"""
Module to add extra functionality like retry, circuitbreaking
to normal requests calls.

read README.md for more info on usage.
"""

import requests
import threading
from circuit_breaker import CircuitBreaker
from config import Config, logger
from urllib.parse import urlparse
from urllib3 import Retry
from requests.exceptions import ConnectionError
from requests import Session
from requests.adapters import HTTPAdapter
from uuid import uuid4

STATUS_FORCELIST = "status_forcelist"
MAX_TRIES = "max_tries"
METHOD_WHITELIST = "method_whitelist"
BACKOFF_FACTOR = "backoff_factor"
TIMEOUT = "timeout"
X_REQUEST_ID = "x-request-id"
HEADERS = "headers"


class Breaker(object):
    """A singleton holding a dictionary of unique key(scheme+netloc+path) to
    a circuitbreaker object.
    """
    _breakers = {}
    _lock = threading.Lock()

    @classmethod
    def get_key(cls, raw_url):
        """creates the key(scheme+netloc+path) from url"""
        o = urlparse(raw_url)
        return "".join([o.scheme, o.netloc, o.path])

    @classmethod
    def new(cls, url):
        """Returs the breaker object. If it exists in cls._breakers, return
        else create a new breaker and save it in cls._breakers before
        returning it.
        """
        key = cls.get_key(url)

        with cls._lock:
            if cls._breakers.get(key):
                logger.info("Using existing circuit for %s", url)
                return cls._breakers[key]

            logger.info("Creating new circuit for %s", url)
            breaker = CircuitBreaker(
                allowed_fails=Config.cb_fail_threshold(),
                retry_time=Config.cb_delay(),
                retry_after=Config.cb_alive_threshold(),
                failure_exceptions=Config.retriable_exceptions())

            cls._breakers[key] = breaker
            return breaker


def get_backoff_args(kwargs, method):
    """Generate backoff arguments"""
    max_tries = kwargs.get(MAX_TRIES, None)
    args = {
        METHOD_WHITELIST: Config.method_whitelist(),
        STATUS_FORCELIST: kwargs.pop(STATUS_FORCELIST,
                                     Config.status_forcelist()),
        BACKOFF_FACTOR: kwargs.pop(BACKOFF_FACTOR,
                                   Config.backoff_factor()),
        MAX_TRIES: kwargs.pop(MAX_TRIES, Config.max_tries())
    }

    if max_tries and max_tries > 0:
        args[METHOD_WHITELIST].append(method.upper())

    args[METHOD_WHITELIST] = frozenset(args[METHOD_WHITELIST])
    return args


def get_timeouts(timeout=None):
    """Ensures that both read and connect timeout are passed in
    final patched requests call. In case any of it is missing in initial
    call by the caller, default values are added"""
    if timeout:
        return timeout
    return (Config.connect_timeout(), Config.read_timeout())


def ensure_request_id(kwargs):
    """Ensures that a unique request-id is present in the request header.
    """
    headers = kwargs.get(HEADERS, {})
    if headers.get(X_REQUEST_ID):
        return
    headers[X_REQUEST_ID] = str(uuid4())
    kwargs[HEADERS] = headers


def patch_pyhystrix(func):
    """Returns a function which takes same arguments as requests.api.request
    and integrates Retry and Circuitbreaking logic in itself before calling
    session.request for actual external api call.
    """
    def pyhystrix_wrapper(method, url, **kwargs):
        logger.info("[%s %s]" % (method, url))
        circuit = Breaker.new(url)
        if circuit.is_open:
            logger.info("OPEN circuit for %s", url)
            circuit.increment_failure_count()
            raise ConnectionError("Open Circuit")

        class CustomRetry(Retry):
            def is_exhausted(self):
                # Increment fail and then check the circuit
                circuit.mark_failure()
                if circuit.is_open:
                    return True
                return super(CustomRetry, self).is_exhausted()

        ensure_request_id(kwargs)
        kwargs[TIMEOUT] = get_timeouts(kwargs.pop(TIMEOUT, None))
        bargs = get_backoff_args(kwargs, method)
        adapter = HTTPAdapter(
            max_retries=CustomRetry(total=bargs[MAX_TRIES],
                                    connect=bargs[MAX_TRIES],
                                    read=bargs[MAX_TRIES],
                                    status=bargs[MAX_TRIES],
                                    status_forcelist=bargs[STATUS_FORCELIST],
                                    method_whitelist=bargs[METHOD_WHITELIST],
                                    backoff_factor=bargs[BACKOFF_FACTOR]))
        s = Session()
        s.mount(url, adapter)
        resp = s.request(method=method, url=url, **kwargs)
        circuit.close()
        return resp

    return pyhystrix_wrapper


def patch_requests():
    """Monkey patching requests.api.request to the wrapper function above using
    patcher. This is not the best way to solve it but for now, this enables us
    to use additional functionality like Retry and circuitbreaking etc with
    least code change at the caller side.
    """
    requests.api.request = patch_pyhystrix(requests.api.request)
    logger.info("pyhystrix added to requests")


def Init():
    """Needs to be called at initialization if the application that will
    use this library.
    """
    patch_requests()



#!/usr/bin/env python

import os
import setuptools
from version import __version__


BASE_DIR = os.path.dirname(os.path.abspath(__file__))
VERSION_FILE = os.path.join(BASE_DIR, 'version.py')
README_FILE = os.path.join(BASE_DIR, "README.md")


# Get the long description from the README file
with open(README_FILE) as f:
    long_description = f.read()


dependencies = [
    "certifi == 2017.4.17",
    "chardet == 3.0.4",
    "httmock == 1.2.6",
    "idna == 2.5",
    "requests == 2.22.0",
    "urllib3 == 1.25.6"
]


setuptools.setup(
    name="pyhystrix",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Mohan Dutt",
    author_email="mohan@oogway.in",
    url="https://github.com/oogway/pyhystrix",
    version=__version__,
    py_modules=['pyhystrix', 'circuit_breaker', 'config', 'version'],
    include_package_data=True,
    install_requires=dependencies,
    test_suite="tests"
)



"""Constants that will be used to define default
behaviour. Can be overridden by setting env vars
"""

import os
import requests
import logging


log_levels = {
    logging.getLevelName(logging.CRITICAL): logging.CRITICAL,
    logging.getLevelName(logging.ERROR): logging.ERROR,
    logging.getLevelName(logging.WARNING): logging.WARNING,
    logging.getLevelName(logging.INFO): logging.INFO,
    logging.getLevelName(logging.DEBUG): logging.DEBUG
}

logger = logging.getLogger("pyhystrix")

level = os.environ.get("PHY_LOG")
logger.setLevel(log_levels.get(level, logging.WARNING))
logger.addHandler(logging.StreamHandler())


class Config(object):
    @staticmethod
    def connect_timeout():
        return int(os.environ.get("PYH_CONNECT_TIMEOUT", 5))

    @staticmethod
    def read_timeout():
        return int(os.environ.get("PYH_READ_TIMEOUT", 5))

    @staticmethod
    def max_tries():
        return int(os.environ.get("PHY_MAX_TRIES", 3))

    @staticmethod
    def backoff_factor():
        return float(os.environ.get("PHY_BACKOFF_FACTOR", 0.5))

    @staticmethod
    def retriable_exceptions():
        return (requests.exceptions.ConnectionError,)

    @staticmethod
    def method_whitelist():
        return ['HEAD', 'GET']

    @staticmethod
    def status_forcelist():
        return [500]

    @staticmethod
    def cb_fail_threshold():
        return int(os.environ.get("PYH_CIRCUIT_FAIL_THRESHOLD", 5))

    @staticmethod
    def cb_alive_threshold():
        return int(os.environ.get("PYH_CIRCUIT_ALIVE_THRESHOLD", 20))

    @staticmethod
    def cb_delay():
        return int(os.environ.get("PYH_CIRCUIT_DELAY", 5))
